/*
  "Fire Emblem 6: Fuuin No Tsurugi" Script Allocator
  By Dark Twilkitri

  Allocates chunks of free memory in the ROM to scripts, and marks
  those areas off so that they cannot be overwritten in the same build.
  Also creates a table for the linker.

  Last modified: 27/09/04
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>

#include "block.h"
#include "link.h"
#include "range.h"


/* Function Prototypes */


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	BLOCK_NFO  curblock;
	char      *enfoname;
	char      *ratname;
	FILE      *enfo;
	FILE      *links;
	FILE      *rat;
	LINK       curlink;
	int        blocklen;
	int        chosenrange;
	int        curarg = 1;
	int        curranges;
	int        enfolen;
	int        i, j;
	int        numblocks;
	int        numranges = 0;
	int        ratlen;
	RANGE     *ranges = 0;

	if(argc < 2)
	{
		printf("alloc usage:\n  alloc script1 [script2] [script3] [...]\n");
		return 1;
	}

	//first collect ranges
	while(curarg < argc)
	{
		//open RAT
		ratname = (char*)malloc(strlen(argv[curarg]) + 5);

		strcpy(ratname, argv[curarg]);
		strcat(ratname, ".rat");

		rat = fopen(ratname, "rb");
		if(!rat)
		{
			//scripts don't _have_ to have an associated RAT.
			curarg++;
			continue;
		}

		//determine number of ranges in file
		fseek(rat, 0, SEEK_END);
		ratlen = ftell(rat);
		fseek(rat, 0, SEEK_SET);

		if((ratlen % sizeof(RANGE)) != 0)
		{
			printf("*** RAT file, \"%s\", is invalid. Skipping...\n", ratname);
			fclose(rat);
			curarg++;
			continue;
		}

		curranges = ratlen / sizeof(RANGE);

		//allocate memory for ranges
		if(!ranges)
			ranges = (RANGE*)malloc(ratlen);
		else
			ranges = (RANGE*)realloc(ranges, (numranges * sizeof(RANGE)) + ratlen);

		if(!ranges)
		{
			printf("Out of available memory. Try closing some programs and trying again.\n");
			fclose(rat);
			return 1;
		}
		//read in ranges
		fread((ranges + numranges), sizeof(RANGE), curranges, rat);

		numranges += curranges;

		free(ratname);
		fclose(rat);
		curarg++;
	}

	//check lowerlimit is lower than upperlimit
	for(i = 0; i < numranges; i++)
	{
		if(ranges[i].lowerlimit <= ranges[i].upperlimit)
			continue;
		//truncate range on error
		printf("Problem with range: lowerlimit %06X, upperlimit %06X.\n", ranges[i].lowerlimit, ranges[i].upperlimit);
		ranges[i].upperlimit = ranges[i].lowerlimit;
	}

	//now ranges should be done, so we can start allocating.
	curarg = 1;

	//open linking file
	links = fopen("fe8e.bln", "wb");
	if(!links)
	{
		printf("*** Could not open target Block Linking file, fe8e.bln.\n");
		return 1;
	}

	while(curarg < argc)
	{
		//open enfo
		enfoname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(enfoname, argv[curarg]);
		strcat(enfoname, ".enfo");

		enfo = fopen(enfoname, "rb");
		if(!enfo)
		{
			printf("Could not open Encoded Information file \"%s\". Skipping...\n", enfoname);
			curarg++;
			continue;
		}

		fseek(enfo, 0, SEEK_END);
		enfolen = ftell(enfo);
		fseek(enfo, 0, SEEK_SET);

		numblocks = enfolen / sizeof(BLOCK_NFO);

		for(i = 0; i < numblocks; i++)
		{
			//read in block
			fread(&curblock, sizeof(BLOCK_NFO), 1, enfo);

			curlink.pointaddress = curblock.pointaddress;

			blocklen = curblock.size;

			//find space for this block
			chosenrange = -1;
			for(j = 0; j < numranges; j++)
			{
				if((ranges[j].upperlimit - ranges[j].lowerlimit + 1) < (unsigned int)blocklen)
					continue;
				if((ranges[j].upperlimit - ranges[j].lowerlimit + 1) == blocklen)
				{
					//exact room required
					chosenrange = j;
					break;
				}
				//else there is enough room
				if(chosenrange == -1)
				{
					//first range with enough room found
					chosenrange = j;
					continue;
				}
				//compare for the best choice (least space left over)
				if((ranges[j].upperlimit - ranges[j].lowerlimit) <
					(ranges[chosenrange].upperlimit - ranges[chosenrange].lowerlimit))
					chosenrange = j;
			}
			if(chosenrange == -1)
			{
				//Was no ranges with enough room
				printf("Not enough room for script %06X.\n", curlink.pointaddress);
				curlink.newaddress = 0xFFFFFFFF;
			}
			else
			{
				//Was enough room in a range - remove this from the range
				curlink.newaddress = ranges[chosenrange].lowerlimit;
				ranges[chosenrange].lowerlimit = ranges[chosenrange].lowerlimit + blocklen;
			}
			//write out link
			fwrite(&curlink, sizeof(LINK), 1, links);


		}

		fclose(enfo);
	
		curarg++;
	}

	fclose(links);
	return 0;
}