/*
  "Fire Emblem 6: Fuuin No Tsurugi" Puretext Huffman Encoder
  By Dark Twilkitri

  Encodes FE6 puretext beneath the Huffman encoding used by the game.

  Last modified: 26/09/04
*/

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>

#include "block.h"
#include "encode.h"


/* Function Prototypes */


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	         BLOCK_NFO  curblock;
			 BLOCK_NFO  newblock;
	         char      *encname;
			 char      *enfoname;
	         char      *infoname;
	         char      *purename;
	unsigned char       curbyte;
	unsigned char       mask;
	         FILE      *bitpatterns;
	         FILE      *encoded;
	         FILE      *enfo;
	         FILE      *info;
	         FILE      *puretext;
			 HUFFLIST  *codetable;
			 int        curarg = 1;
	unsigned int        curoff = 0;
	unsigned int        donelen = 0;
	         int        i, j;
             int        numcodes;
	unsigned int        purelen;
			 int        tbllen;
	unsigned short      curchar;

	if(argc < 2)
	{
		printf("encode usage:\n  encode script1 [script2] [script3] [...]\n");
		return 1;
	}

	//open bitpattern table
	bitpatterns = fopen("fe8e.bit", "rb");
	if(!bitpatterns)
	{
		printf("** Could not open Bit Pattern table, fe8e.bit.\nPlease check that this file is available; cannot encode without it.\n");
		return 1;
	}

	//get length of table
	fseek(bitpatterns, 0, SEEK_END);
    tbllen = ftell(bitpatterns);
	fseek(bitpatterns, 0, SEEK_SET);

	//allocate memory to store table
	if((tbllen % sizeof(HUFFLIST)) != 0)
	{
		//table not valid; must be integer amount of entries
		printf("** The Bit Pattern table, fe8e.bit, is not valid.\nPlease obtain a valid one; cannot encode without it.\n");
		fclose(bitpatterns);
		return 1;
	}
	codetable = (HUFFLIST*)malloc(tbllen);
	if(!codetable)
	{
		printf("** Not enough memory available to store Bit Pattern table.\nPlease close some running programs and try again.\n");
		fclose(bitpatterns);
		return 1;
	}

	//read in table
	numcodes = tbllen / sizeof(HUFFLIST);
	fread(codetable, sizeof(HUFFLIST), numcodes, bitpatterns);

	//now that the table is in memory, don't need the file open any more; less cleaning to do later
	fclose(bitpatterns);

	while(curarg < argc)
	{
		//open puretext and information files
		purename = (char*)malloc(strlen(argv[curarg]) + 6);
		//assume enough room

		strcpy(purename, argv[curarg]);
		strcat(purename, ".dte");
		
		//open puretext
		puretext = fopen(purename, "rb");
		if(!puretext)
		{
			printf("*** Could not open script puretext file, \"%s\". Skipping script...\n", purename);
			curarg++;
			continue;
		}

		//get length
		fseek(puretext, 0, SEEK_END);
		purelen = ftell(puretext);
		fseek(puretext, 0, SEEK_SET);

		infoname = (char*)malloc(strlen(argv[curarg]) + 6);

		strcpy(infoname, argv[curarg]);
		strcat(infoname, ".info");

		//open information
		info = fopen(infoname, "rb");
		if(!info)
		{
			printf("*** Could not open script information file, \"%s\". Skipping script...\n", infoname);
			fclose(puretext);
			curarg++;
			continue;
		}

		//create target files
		encname = (char*)malloc(strlen(argv[curarg]) + 5);

		strcpy(encname, argv[curarg]);
		strcat(encname, ".enc");

		//open encoded file
		encoded = fopen(encname, "wb");
		if(!encoded)
		{
			printf("*** Could not open target Encoded file, \"%s\". Skipping script...\n", encname);
			fclose(puretext);
			fclose(info);
			curarg++;
			continue;
		}

		enfoname = (char*)malloc(strlen(argv[curarg]) + 6);

		strcpy(enfoname, argv[curarg]);
		strcat(enfoname, ".enfo");

		//open encoded information file
		enfo = fopen(enfoname, "wb");
		if(!enfo)
		{
			printf("*** Could not open target Encoded Information file, \"%s\". Skipping script...\n", enfoname);
			fclose(puretext);
			fclose(info);
			fclose(encoded);
			curarg++;
			continue;
		}

		while(donelen < purelen)
		{
			//load block information
			fread(&curblock, sizeof(BLOCK_NFO), 1, info);

			//start new block information
			newblock.pointaddress = curblock.pointaddress;
			newblock.size = 0;

			//printf("%06X %06X %06X %06X\n", curblock.pointaddress, curblock.size, donelen, purelen);
			curbyte = 0;
			mask = 0x01;
			while(curoff < (donelen + curblock.size))
			{
				//ALL TEXT CODES ARE 2 BYTES WIDE. No checking is necessary.
				fread(&curchar, 2, 1, puretext);
				//printf("\n%04X\n", curchar);
				curoff += 2;

				//find character in table
				for(i = 0; i < numcodes; i++)
					if(codetable[i].character == curchar)
						break;
				if(i == numcodes)
				{
					//something went wrong
					printf("* Warning: Character \'%04X\' not found in Bit Pattern table.\n", curchar);
					continue;
				}
				for(j = 0; j < (int)strlen(codetable[i].code); j++)
				{
					if(codetable[i].code[j] == '1')
					//{
						curbyte += (0xFF & mask);
						//printf("%02X\t", curbyte);
					//}
					mask = mask << 1;
					if(mask == 0)
					{
						//this byte is done
						fputc(curbyte, encoded);
						newblock.size++;
						curbyte = 0;
						mask = 0x01;
					}
				}

			}
			//add in what's left over in the final byte
			if(mask != 0x01)
			{
				fputc(curbyte, encoded);
				newblock.size++;
			}

			//keep track of the current completed blocks
			donelen += curblock.size;

			//write out the block information
			fwrite(&newblock, sizeof(BLOCK_NFO), 1, enfo);
		}
		
		//close files
		free(purename);
		free(infoname);
		free(encname);
		free(enfoname);
		fclose(puretext);
		fclose(info);
		fclose(encoded);
		fclose(enfo);
		curarg++;
		curoff = 0;
		donelen = 0;
	}

	//Kill Program
	return 0;
}