/*
  "Fire Emblem 6: Fuuin No Tsurugi" IPS Creator
  By Dark Twilkitri

  This program will create an IPS patch which overwrites the allocated areas
  with their new text, and relinks the pointers.

  Last Modified: 27/09/04
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>

#include "block.h"
#include "link.h"


/* Function Prototypes */


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	BLOCK_NFO *blocklist;
	char      *block;
	char      *encname;
	char      *enfoname;
	FILE      *links;
	FILE      *encoded = NULL;
	FILE      *enfo;
	FILE      *ips;
	int        curarg;
	int        curblocks;
	int        enclen;
	int        i, j;
	int        numblocks = 0;
	int        numlinks;
	LINK       curlink;
	LINK      *linklist;

	if(argc < 2)
	{
		printf("link usage:\n  link script1 [script2] [script3] [...]\n");
		return 1;
	}

	//first order of business is to collect all the block information
	curarg = 1;
	while(curarg < argc)
	{
		//open Encoded Information file
		enfoname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(enfoname, argv[curarg]);
		strcat(enfoname, ".enfo");

		enfo = fopen(enfoname, "rb");
		if(!enfo)
		{
			printf("*** Could not open script Encoded Information file, \"%s\". Skipping...\n", enfoname);
			//don't want text for this, if block is non-existant
			argv[curarg][0] = '\0';
			free(enfoname);
			curarg++;
			continue;
		}

		fseek(enfo, 0, SEEK_END);
		curblocks = ftell(enfo) / sizeof(BLOCK_NFO);
		fseek(enfo, 0, SEEK_SET);

		if(numblocks == 0)
			blocklist = (BLOCK_NFO*)malloc(curblocks * sizeof(BLOCK_NFO));
		else
			blocklist = (BLOCK_NFO*)realloc(blocklist, (numblocks + curblocks) * sizeof(BLOCK_NFO));

		//read in blocks
		fread(&(blocklist[numblocks]), sizeof(BLOCK_NFO), curblocks, enfo);
		numblocks += curblocks;

		free(enfoname);
		fclose(enfo);
		curarg++;
	}

	//allocate memory for link list
	linklist = (LINK*)malloc(numblocks * sizeof(LINK));

	//read in Block Links and associate with blocks
	links = fopen("fe8e.bln", "rb");
	if(!links)
	{
		printf("*** Could not open Block Linking file, \"fe8e.bln\". Can not link without it.\n");
		return 1;
	}

	fseek(links, 0, SEEK_END);
	numlinks = ftell(links) / sizeof(LINK);
	fseek(links, 0, SEEK_SET);

	for(i = 0; i < numlinks; i++)
	{
		fread(&curlink, sizeof(LINK), 1, links);
		for(j = 0; j < numblocks; j++)
		{
			if(curlink.pointaddress == blocklist[j].pointaddress)
				break;
		}

		if(j == numblocks) //no matching block was found
		{
			printf("Link %06X missing block information. Discarded...\n", curlink.pointaddress);
			continue;
		}
		else
		{
			//insert link at same index
			linklist[j].pointaddress = curlink.pointaddress;
			linklist[j].newaddress = curlink.newaddress;
		}
	}

	//close links file
	fclose(links);

	//open ips
	ips = fopen("fe8e.ips", "wb");
	if(!ips)
	{
		printf("Could not open target IPS file.\n");
		return 1;
	}
	//patch header
	fputc('P', ips);
	fputc('A', ips);
	fputc('T', ips);
	fputc('C', ips);
	fputc('H', ips);

	//loop through blocks and encoded text files
	curarg = 1;
	for(i = 0; i < numblocks; i++)
	{
		if(!encoded)
		{
			//no encoded script currently open; open the next one
			//check that we haven't nulled the script out (no enfo file) first
			while((argv[curarg][0] == '\0') & (curarg < argc))
				curarg++;
			if(curarg > argc)
				break;        //out of encoded scripts

			encname = (char*)malloc(strlen(argv[curarg]) + 5);
			encname[0] = '\0';
			strcpy(encname, argv[curarg]);
			strcat(encname, ".enc");

			encoded = fopen(encname, "rb");
			if(!encoded)
			{
				//encoded script is missing; irrecoverable error
				printf("*** Missing encoded script \"%s\". Unable to continue; shutting down.\n", encname);
				fclose(ips);
				return 1;
			}

			fseek(encoded, 0, SEEK_END);
			enclen = ftell(encoded);
			fseek(encoded, 0, SEEK_SET);
		}

		//allocate room for block
		block = (char*)malloc(blocklist[i].size);

		//read in block
		fread(block, 1, blocklist[i].size, encoded);

		if((ftell(encoded) + 1) >= enclen)
		{
			//done with this file - will open new one at start of next iteration
			fclose(encoded);
			curarg++;
			encoded = NULL;
		}

		if(linklist[i].newaddress == 0xFFFFFFFF)
		{
			//there was no room for this block (still had to read it in, though)
			free(block);
			continue;
		}

		//update block link
		//pointer pointer
		fputc((blocklist[i].pointaddress & 0xFF0000) >> 16, ips);
		fputc((blocklist[i].pointaddress & 0x00FF00) >> 8, ips);
		fputc((blocklist[i].pointaddress & 0x0000FF), ips);
		//size (always 3)
		fputc(0x00, ips);
		fputc(0x03, ips);
		//new pointer (in reverse)
		fputc((linklist[i].newaddress & 0x0000FF), ips);
		fputc((linklist[i].newaddress & 0x00FF00) >> 8, ips);
		fputc((linklist[i].newaddress & 0xFF0000) >> 16, ips);

		//write block
		//block pointer
		fputc((linklist[i].newaddress & 0xFF0000) >> 16, ips);
		fputc((linklist[i].newaddress & 0x00FF00) >> 8, ips);
		fputc((linklist[i].newaddress & 0x0000FF), ips);
		//size (assume small enough, currently)
		fputc((blocklist[i].size & 0xFF00) >> 8, ips);
		fputc((blocklist[i].size & 0x00FF), ips);
		//encoded text
		fwrite(block, 1, blocklist[i].size, ips);

		free(block);
		free(encname);
	}
	return 0;
}