/*
 * "Fire Emblem" Dual Tile Encoding setup
 * By Tristan Wilkinson
 *
 * DEPRECATED -> Takes a parsed (pure) script file and measures frequencies of all
 * valid letter pairs.
 * Creates a list of valid shorts consisting of control codes,
 * single characters, and the most frequent pairs. <- DEPRECATED
 * Reads in the normal list control codes, single characters, and pairs.
 * Then rewrites the file so that it uses only shorts from this list.
 *
 * Last modified: 05/02/05
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "block.h"
#include "concrete.h"
#include "frequency.h"

#define  CHARS_START 0x15A72C
#define  CHARS_END   0x15BDDB

/* Function Prototypes */
int freqcomp(const void* freq1, const void* freq2);


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	BLOCK_NFO  curblock;
	BLOCK_NFO  newblock;
	char      *dtename;
	char      *infoname;
	char      *purename;
	FILE      *dte;
	FILE      *inforead;
	FILE      *infowrite;
	FILE      *pure;
	FILE      *rom;
	FREQUENCY *freqtbl;
	int        curallow;
	int        curarg = 1;
	int        curfreq;
	int        curshorhi;
	int        curshorlo;
	int        curoff;
	int        curpair;
	int        donelen;
	int        i;
	int        maxallows = 0;
	int        maxfreqs = 100;
	int        numallows;
	int        numfreqs = 0;
	int        purelen;
	short      allowlist[1452];

	//check arguments
	if(argc < 2)
	{
		fprintf(stderr, "optidte usage:\n  optidte script1 [script2] [script3] ...\n");
		return 1;
	}

	freqtbl = (FREQUENCY*)malloc(maxfreqs * sizeof(FREQUENCY));

	/*** DEPRECATED ***/
	//first collect frequency information from all scripts
	while(curarg < argc)
	{
		//open current puretext
		purename = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(purename, argv[curarg]);
		strcat(purename, ".pure");

		pure = fopen(purename, "rb");
		if(!pure)
		{
			fprintf(stderr, "Could not open puretext file %s. Skipping...\n", purename);
			free(purename);
			curarg++;
			continue;
		}
		free(purename);

		//get length
		fseek(pure, 0, SEEK_END);
		purelen = ftell(pure);
		fseek(pure, 0, SEEK_SET);

		curoff = 0;

		//get frequencies
		while((curoff + 3) < purelen)
		{
			//reset file pointer
			fseek(pure, curoff, SEEK_SET);

			curshorlo = fgetc(pure);
			curshorlo += fgetc(pure) << 8;
			//check for command
			if((curshorlo & 0x0000FF00) != 0x1F00)
			{
				//command
				//so we skip this pair
				curoff += 2;
				if((curshorlo & 0x000000FF) == 0x10) //openchar takes an argument
					curoff += 2;
				continue;
			}
			curshorhi = fgetc(pure);
			curshorhi += fgetc(pure) << 8;
			//check for command
			if((curshorhi & 0x0000FF00) != 0x1F00)
			{
				//command
				//so we skip this pair and the last one
				curoff += 4;
				if((curshorhi & 0x000000FF) == 0x10) //openchar takes an argument
					curoff += 2;
				continue;
			}

			//if we get here we should have a pair of characters
			//put them together
			curpair = (curshorhi & 0x000000FF);
			curpair += (curshorlo & 0x000000FF) << 8;

			//increase offset
			curoff += 4;

			//now search for this pair in the frequency table
			for(i = 0; i < numfreqs; i++)
			{
				if(freqtbl[i].pair == curpair)
				{
					freqtbl[i].numoccur++;
					break;
				}
			}

			//if i isn't equal to numfreqs, we found it
			if(i != numfreqs)
				continue;

			//create new entry
			//first check for space
			if(numfreqs == maxfreqs)
			{
				maxfreqs += 100;
				freqtbl = (FREQUENCY*)realloc(freqtbl, maxfreqs * sizeof(FREQUENCY));
			}
			
			//now add frequency entry
			freqtbl[numfreqs].pair = curpair;
			freqtbl[numfreqs].numoccur = 1;

			numfreqs++;

			//continue
		}

		//close puretest
		fclose(pure);

		//next puretext
		curarg++;
	}
	
	maxfreqs = numfreqs;

	//now we should have collected all frequency information
	//sort it
	qsort(freqtbl, numfreqs, sizeof(FREQUENCY), freqcomp);

	/*//test stuff
	for(i = 0; i < numfreqs; i++)
	{
		printf("%c%c - %d\n", (freqtbl[i].pair & 0xFF00) >> 8, freqtbl[i].pair & 0x00FF, freqtbl[i].numoccur);
	}*/

	//create allowable list
	memset(allowlist, 0, 1452 * sizeof(short));

	numallows = 0;

	//populate allowlist with concretes
	while(numallows * 4 < 532)
	{
		allowlist[numallows] = concretes[numallows * 4] + (concretes[(numallows * 4) + 1] << 8);
		numallows++;
	}

	//now fill in the rest with the most popular pairs
	curfreq = 0;
	while(numallows < 1452)
	{
		allowlist[numallows] = freqtbl[curfreq].pair;
		numallows++;
		curfreq++;
		if(curfreq == maxfreqs) //ran out of frequencies to add...
		{
			maxallows = numallows;
			break;
		}
	}
	
/*	//read in character list from ROM
	//open ROM (assume fe8e.gba)
	rom = fopen("fe8e.gba", "rb");
	if(!rom)
	{
		printf("Could not open FE* ROM, \"fe8e.gba\"; need this for character information.\n");
		return 1;
	}

	//collect character information
	fseek(rom, CHARS_START, SEEK_SET);

	while(ftell(rom) < CHARS_END)
	{
		fread(&(allowlist[maxallows++]), sizeof(short), 1, rom);
		fgetc(rom); fgetc(rom);
	}

	//close rom
	fclose(rom);*/

	//now the allowable list should be full
	//go back to the first puretext
	curarg = 1;

	//we will now convert the puretexts into dte-optimized puretexts
	while(curarg < argc)
	{
		//open current puretext
		purename = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(purename, argv[curarg]);
		strcat(purename, ".pure");

		pure = fopen(purename, "rb");
		if(!pure)
		{
			fprintf(stderr, "Could not open puretext file %s. Skipping...\n", purename);
			free(purename);
			curarg++;
			continue;
		}

		free(purename);

		//get length
		fseek(pure, 0, SEEK_END);
		purelen = ftell(pure);
		fseek(pure, 0, SEEK_SET);

		//open current infomation
		infoname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(infoname, argv[curarg]);
		strcat(infoname, ".info");

		inforead = fopen(infoname, "rb");
		if(!inforead)
		{
			fprintf(stderr, "Could not open information file %s. Skipping...\n", infoname);
			free(infoname);
			fclose(pure);
			curarg++;
			continue;
		}

		infowrite = fopen(infoname, "r+b");
		if(!infowrite)
		{
			fprintf(stderr, "Could not open information file %s. Skipping...\n", infoname);
			free(infoname);
			fclose(pure);
			fclose(inforead);
			curarg++;
			continue;
		}

		free(infoname);

		//open dte output file
		dtename = (char*)malloc(strlen(argv[curarg]) + 5);
		strcpy(dtename, argv[curarg]);
		strcat(dtename, ".dte");

		dte = fopen(dtename, "wb");
		if(!dte)
		{
			fprintf(stderr, "Could not open dte file %s. Skipping...\n", dtename);
			free(dtename);
			fclose(pure);
			fclose(inforead);
			fclose(infowrite);
			curarg++;
			continue;
		}

		free(dtename);

		curoff = 0;
		donelen = 0;

		while(donelen < purelen)
		{
			//load block information
			fread(&curblock, sizeof(BLOCK_NFO), 1, inforead);

			//initialize new block
			newblock.pointaddress = curblock.pointaddress;
			newblock.size = 0;

			while((unsigned)(curoff + 3) <= (donelen + curblock.size))
			{
				//reset file pointer
				fseek(pure, curoff, SEEK_SET);

				curshorlo = fgetc(pure);
				curshorlo += fgetc(pure) << 8;
				//check for command
				if((curshorlo & 0x0000FF00) != 0x1F00)
				{
					//command
					//write the literal command
					fputc(curshorlo & 0x000000FF, dte);
					fputc((curshorlo & 0x0000FF00) >> 8, dte);
					curoff += 2;
					newblock.size += 2;
					if((curshorlo & 0x000000FF) == 0x10) //openchar takes an argument
					{
						curshorlo = fgetc(pure);
						curshorlo += fgetc(pure) << 8;
						fputc(curshorlo & 0x000000FF, dte);
						fputc((curshorlo & 0x0000FF00) >> 8, dte);
						curoff += 2;
						newblock.size += 2;
					}
					continue;
				}
				curshorhi = fgetc(pure);
				curshorhi += fgetc(pure) << 8;
				//check for command
				if((curshorhi & 0x0000FF00) != 0x1F00)
				{
					//command
					//so we write the concrete for the last letter and the literal command
					fputc(curshorlo & 0x000000FF, dte);
					fputc((curshorlo & 0x0000FF00) >> 8, dte);
					fputc(curshorhi & 0x000000FF, dte);
					fputc((curshorhi & 0x0000FF00) >> 8, dte);
					curoff += 4;
					newblock.size += 4;
					if((curshorhi & 0x000000FF) == 0x10) //openchar takes an argument
					{
						curshorhi = fgetc(pure);
						curshorhi += fgetc(pure) << 8;
						fputc(curshorhi & 0x000000FF, dte);
						fputc((curshorhi & 0x0000FF00) >> 8, dte);
						curoff += 2;
						newblock.size += 2;
					}
					continue;
				}

				//if we get here we should have a pair of characters
				//put them together
				curpair = (curshorhi & 0x000000FF);
				curpair += (curshorlo & 0x000000FF) << 8;

				//now, look for this pair in the allowlist
				curallow = 0;
				while(curallow < maxallows)
				{
					if(allowlist[curallow++] == curpair)
						break;
				}

				if(curallow == maxallows) //did not find a pair
				{
					//write out the concrete for the lo character
					//and shift along one
					fputc(curshorlo & 0x000000FF, dte);
					fputc((curshorlo & 0x0000FF00) >> 8, dte);
					curoff += 2;
					newblock.size += 2;
					continue;
				}
				else //did find a pair
				{
					//write out the pair
					//and shift along four
					fputc(curshorlo & 0x000000FF, dte);
					fputc(curshorhi & 0x000000FF, dte);
					curoff += 4;
					newblock.size += 2;
					continue;
				}

			}
			//check that there's not still a character remaining
			if((unsigned)(curoff) < (donelen + curblock.size))
			{
				//write this out as a literal
				fseek(pure, curoff, SEEK_SET);

				fputc(fgetc(pure), dte);
				fputc(fgetc(pure), dte);
				curoff += 2;
				newblock.size += 2;
			}

			//update done length
			donelen += curblock.size;

			//write out new block information
			fwrite(&newblock, sizeof(BLOCK_NFO), 1, infowrite);
		}

		//check that there's not still a character remaining
		if((curoff + 1) < purelen)
		{
			//write this out as a literal
			fseek(pure, curoff, SEEK_SET);

			fputc(fgetc(pure), dte);
			fputc(fgetc(pure), dte);
		}

		//close files
		fclose(pure);
		fclose(inforead);
		fclose(infowrite);
		fclose(dte);

		//next arg
		curarg++;

		//continue
	}

}


int freqcomp(const void* freq1, const void* freq2)
{
	return (((FREQUENCY*)freq2)->numoccur) - (((FREQUENCY*)freq1)->numoccur);
}