/*
  "Fire Emblem 6: Fuuin No Tsurugi" Script Parser
  By Dark Twilkitri
    Overriden to work with "Fire Emblem" (Fire Emblem 7)
	  Overridden to work with "Fire Emblem: The Sacred Stones" (Fire Emblem 8)

  Converts text-based scripts to the natural codes used by the game.

  Last Modified: 04/02/05
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "block.h"

/* Function Prototypes */


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	BLOCK_NFO  block;
	char      *argument;
	char      *command;
	char      *infoname;
	char      *parsedname;
	char       curchar;
	char       curdirchar;
	char       comment[8192];
	char       directive[100];
	FILE      *info;
	FILE      *parsed;
	FILE      *script;
	int        blklen;
	int        curarg = 1;        //current script
	int        curoff;
	int        directarg;
	int        i;
	int        scrlen;

	//parse arguments
	if(argc < 2)
	{
		printf("parse usage:\n  parse script1 [script2] [script3] ...");
		return 1;
	}

	while(curarg < argc)
	{
		//open script for parsing
		script = fopen(argv[curarg], "rb");
		if(!script)
		{
			fprintf(stderr, "*** Unable to open script \"%s\"\n", argv[curarg]);
			curarg++;
			continue;
		}

		parsedname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(parsedname, argv[curarg]);
		strcat(parsedname, ".pure");
		parsed = fopen(parsedname, "wb");
		if(!parsed)
		{
			fprintf(stderr, "*** Unable to open parser target \"%s\"\n", parsedname);
			fclose(script);
			curarg++;
			continue;
		}

		infoname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(infoname, argv[curarg]);
		strcat(infoname, ".info");
		info = fopen(infoname, "wb");
		if(!info)
		{
			fprintf(stderr, "*** Unable to open parser target \"%s\"\n", infoname);
			fclose(script);
			fclose(parsed);
			curarg++;
			continue;
		}

		//get length
		fseek(script, 0, SEEK_END);
		scrlen = ftell(script);
		fseek(script, 0, SEEK_SET);

		curoff = 0;

		block.pointaddress = 0xFFFFFFFF;
		block.size = 0;
		blklen = 0;

		//iterate through script
		while(curoff < scrlen)
		{
			//get character
			curchar = fgetc(script);

			if(curchar == '#')
			{
				//comment - go to end of line
				fgets(comment, 8192, script);
				curoff += (int)strlen(comment) + 1;
				for(i = 0; i < (int)strlen(comment); i++)
				{
					if(comment[i] != '@')
						continue;
					if(comment[i + 1] != '@')
						continue;
					//pointer address
					block.pointaddress = strtol(&(comment[i + 2]), NULL, 16);
					break;
				}
			}
            else if(curchar == '[')
			{
				curdirchar = 0;
				//directive
				while((directive[curdirchar] = fgetc(script)) != ']')
					curdirchar++;
				curdirchar++;
				directive[curdirchar] = '\0';

				//get command
				command = strtok(directive, " ]");
				_strlwr(command);
				if(!strcmp(command, "x"))
				{
					fputc(0x00, parsed);
					fputc(0x00, parsed);
					block.size = blklen + 2;
					fwrite(&block, sizeof(block), 1, info);
					block.pointaddress = 0xFFFFFFFF;
					block.size = 0;
					blklen = 0;
				}
				else if(!strcmp(command, "nl"))
				{
					fputc(0x01, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "k"))
				{
					fputc(0x03, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targfarleft"))
				{
					fputc(0x08, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targmedileft"))
				{
					fputc(0x09, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targcloseleft"))
				{
					fputc(0x0A, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targcloseright"))
				{
					fputc(0x0B, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targmediright"))
				{
					fputc(0x0C, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "targfarright"))
				{
					fputc(0x0D, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "openchar"))
				{
					fputc(0x10, parsed);
					fputc(0x00, parsed);
					argument = strtok(NULL, " ]");
					_strlwr(argument);
					if(!strcmp(argument, "curchar"))
					{
						fputc(0xFF, parsed);
						fputc(0xFF, parsed);
					}
					else
					{
						directarg = strtol(argument, NULL, 16);
						fputc(directarg & 0x00FF, parsed);
						fputc((directarg & 0xFF00) >> 8, parsed);
					}
					blklen += 4;
				}
				else if(!strcmp(command, "exitchar"))
				{
					fputc(0x11, parsed);
					fputc(0x00, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "ctrl"))
				{
					argument = strtok(NULL, " ]");
					directarg = strtol(argument, NULL, 16);
					fputc(directarg & 0x00FF, parsed);
					fputc((directarg & 0xFF00) >> 8, parsed);
					blklen += 2;
				}
				curoff += curdirchar + 1;
			}
			else if((curchar == '\r') || (curchar == '\n'))
			{
				//ignore
				curoff++;
			}
			else
			{
				//make all text single for the momemnt
				fputc(curchar, parsed);
				fputc(0x1F, parsed);
				curoff++;
				blklen += 2;
			}
		}
		//curchar = fgetc(script);
		

		//free memory
		free(parsedname);
		free(infoname);

		//close script
		fclose(info);
		fclose(parsed);
		fclose(script);

		//next script
		curarg++;
	}

	return 0;
}