/*
  "Fire Emblem 6: Fuuin No Tsurugi" Puretext Script Parser
  By Dark Twilkitri

  Converts text-based puretext-type scripts to the natural codes used by the game.

  Last Modified: 14/12/04
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "block.h"
#include "engcodes.h"

#define  LANG_JAP 1
#define  LANG_ENG 0

/* Function Prototypes */


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	BLOCK_NFO  block;
	char      *argument;
	char      *command;
	char      *infoname;
	char      *parsedname;
	char       curcharlo;
	char       curcharhi;
	char       curdirchar;
	char       comment[8192];
	char       directive[100];
	FILE      *enfo;
	FILE      *parsed;
	FILE      *script;
	int        blklen;
	int        curarg = 1;        //current script
	int        curoff;
	int        directarg;
	int        i;
	int        language = LANG_JAP;
	int        scrlen;

	//parse arguments
	if(argc < 2)
	{
		printf("parse usage:\n  parse script1 [script2] [script3] ...");
		return 1;
	}

	while(curarg < argc)
	{
		//open script for parsing
		script = fopen(argv[curarg], "rb");
		if(!script)
		{
			fprintf(stderr, "*** Unable to open script \"%s\"\n", argv[curarg]);
			curarg++;
			continue;
		}

		parsedname = (char*)malloc(strlen(argv[curarg]) + 5);
		strcpy(parsedname, argv[curarg]);
		strcat(parsedname, ".enc");
		parsed = fopen(parsedname, "wb");
		if(!parsed)
		{
			fprintf(stderr, "*** Unable to open parser target \"%s\"\n", parsedname);
			fclose(script);
			curarg++;
			continue;
		}

		infoname = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(infoname, argv[curarg]);
		strcat(infoname, ".enfo");
		enfo = fopen(infoname, "wb");
		if(!enfo)
		{
			fprintf(stderr, "*** Unable to open parser target \"%s\"\n", infoname);
			fclose(script);
			fclose(parsed);
			curarg++;
			continue;
		}

		//get length
		fseek(script, 0, SEEK_END);
		scrlen = ftell(script);
		fseek(script, 0, SEEK_SET);

		curoff = 0;

		block.pointaddress = 0xFFFFFFFF;
		block.size = 0;
		blklen = 0;

		//iterate through script
		while(curoff < scrlen)
		{
			//get first byte of character
			curcharlo = fgetc(script);

			//assume everything we don't recognise is a japanese character.
			if(curcharlo == '#')
			{
				//comment - go to end of line
				fgets(comment, 8192, script);
				curoff += (int)strlen(comment) + 1;
				for(i = 0; i < (int)strlen(comment); i++)
				{
					if(comment[i] != '@')
						continue;
					if(comment[i + 1] != '@')
						continue;
					//pointer address
					block.pointaddress = strtol(&(comment[i + 2]), NULL, 16);
					break;
				}
			}
            else if(curcharlo == '[')
			{
				curdirchar = 0;
				//directive
				while((directive[curdirchar] = fgetc(script)) != ']')
					curdirchar++;
				curdirchar++;
				directive[curdirchar] = '\0';

				//get command
				/* PUREPARSER NOTE
				   [x] is the only game command we will listen to here
				       apart from [ctrl ...]
				   [ascii] and [sjis] directives are still applicable
				   */
				command = strtok(directive, " ]");
				_strlwr(command);
				if(!strcmp(command, "x"))
				{
					fputc(0x00, parsed);
					fputc(0x00, parsed);
					block.size = blklen + 2;
					fwrite(&block, sizeof(block), 1, enfo);
					block.pointaddress = 0xFFFFFFFF;
					block.size = 0;
					blklen = 0;
				}
				else if(!strcmp(command, "ctrl"))
				{
					argument = strtok(NULL, " ]");
					directarg = strtol(argument, NULL, 16);
					fputc(directarg & 0x00FF, parsed);
					fputc((directarg & 0xFF00) >> 8, parsed);
					blklen += 2;
				}
				else if(!strcmp(command, "ascii"))
					language = LANG_ENG;
				else if(!strcmp(command, "sjis"))
					language = LANG_JAP;
				curoff += curdirchar + 1;
			}
			else if((curcharlo == '\r') || (curcharlo == '\n'))
			{
				//ignore
				curoff++;
			}
			else if((curcharlo == ' ') && (language == LANG_JAP))
			{
				fputc(0x81, parsed);
				fputc(0x40, parsed);
				curoff++;
				blklen += 2;
			}
			else
			{
				if(language == LANG_JAP)
				{
					curcharhi = fgetc(script);
					fputc(curcharlo, parsed);
					fputc(curcharhi, parsed);
					curoff += 2;
					blklen += 2;
				}
				else //LANG_ENG
				{
					fputc(english[0][curcharlo], parsed);
					fputc(english[1][curcharlo], parsed);
					curoff++;
					blklen += 2;
				}
			}
		}
		//curchar = fgetc(script);
		

		//free memory
		free(parsedname);
		free(infoname);

		//close script
		fclose(enfo);
		fclose(parsed);
		fclose(script);

		//next script
		curarg++;
	}

	return 0;
}