/*
  "Fire Emblem 6: Fuuin No Tsurugi" Huffman Table Generator
  By Dark Twilkitri

  Scans the entire puretext to be inserted into the game, and generates
  an efficient Huffman Table for insertion into the game.
  Also creates a matching Bit Pattern table for use with the encoder.

  Last Modified: 28/09/04
 */

/* Preprocessor */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "node.h"

#define  CHARS_START 0x15A72C
#define  CHARS_END   0x15BDDB

#define  NUM_CHARS   0x05AC

/* Function Prototypes */
int  freqcomp(const void* freq1, const void* freq2);
void createtable(NODE *encoded, int numnodes);


/* Globals */


/* Main */
int main(int argc, char* argv[])
{
	/* Locals */
	char      *purename;          //puretext file's name
	FILE      *puretext;          //to collect frequency information
	FILE      *enctable;          // 'fake' encoded & enformation files
	//FILE      *rom;
	FREQUENCY  freqtable[8192];   //table of frequencies for each character
	FREQUENCY  newfreq;           //new frequency
	int        curarg = 1;        //current script being analysed
	int        curchar;           //current character
	int        filler = 0x26;     //filler for remaining table slots
	int        i;                 //generic iterator
	int        numfreqs = 0;      //number of entries in frequency table
	int        numnodes = 0;      //number of entries in huffman table
	int        purelen;           //length of puretext file
	NODE       hufftable[8192];   //huffman table

	//check arguments
	if(argc < 2)
	{
		printf("retable usage:\n  retable script1 [script2] [script3] [...]");
		return 1;
	}

/*	//open ROM (assume fe8e.gba)
	rom = fopen("fe8e.gba", "rb");
	if(!rom)
	{
		printf("Could not open FE8 ROM, \"fe8e.gba\"; need this for character information.\n");
		return 1;
	}

	//collect character information
	fseek(rom, CHARS_START, SEEK_SET);

	while(ftell(rom) < CHARS_END)
		fread(&(freqtable[numfreqs++].node), sizeof(NODE), 1, rom);

	//close rom
	fclose(rom);*/

	//gather frequency information
	while(curarg < argc)
	{
		//open script's puretext
		purename = (char*)malloc(strlen(argv[curarg]) + 6);
		strcpy(purename, argv[curarg]);
		strcat(purename, ".dte");

		puretext = fopen(purename, "rb");
		if(!puretext)
		{
			printf("Could not open puretext file for script \"%s\". Skipping...\n", argv[curarg]);
			free(purename);
			curarg++;
			continue;
		}

		//collect characters and increment their frequency
		fseek(puretext, 0, SEEK_END);
		purelen = ftell(puretext);
		fseek(puretext, 0, SEEK_SET);

		while(ftell(puretext) < purelen)
		{
			curchar = fgetc(puretext);
			curchar += fgetc(puretext) << 8;

			//find character in frequency table
			for(i = 0; i < numfreqs; i++)
				if(freqtable[i].node.lname == curchar)
					break;

			if(i != numfreqs)
			{
				//increment frequency of character
				freqtable[i].numoccur++;
			}
			else
			{
				//add character to table
				freqtable[i].node.lname = (unsigned short)curchar;
				freqtable[i].node.rname = (unsigned short)0xFFFF;
				freqtable[i].numoccur = 1;
				freqtable[i].level = 0;
				numfreqs++;
				/*//error - character not in table
				fprintf(stderr, "* Character(s) %04X missing from frequency table.\n", curchar);*/
			}
				
		}

		//fill out table on lower / optidte should enforce the maximum limit
		while(numfreqs < 0x5AC)
		{
			freqtable[numfreqs].node.lname = (unsigned short)filler++;
			freqtable[numfreqs].node.rname = (unsigned short)0xFFFF;
			freqtable[numfreqs].numoccur = 0;
			freqtable[numfreqs].level = 0;
			numfreqs++;
		}


		//clean up
		fclose(puretext);
		free(purename);

		curarg++;
	}

	//copy character nodes into start of hufftable
	for(i = 0; i < numfreqs; i++)
	{
		hufftable[i].lname = freqtable[i].node.lname;
		hufftable[i].rname = freqtable[i].node.rname;
		numnodes++;
	}

	//all frequency information should now be collected
	//build Huffman Table until we only have 1 entry left
	while(numfreqs > 1)
	{
		//sort frequency table
		qsort(freqtable, numfreqs, sizeof(FREQUENCY), freqcomp);

		//create new frequency combining lowest two frequencies
		newfreq.numoccur = freqtable[numfreqs - 1].numoccur + freqtable[numfreqs - 2].numoccur;
		
		//determine level of node
		if(freqtable[numfreqs - 1].level > freqtable[numfreqs - 2].level)
			newfreq.level = freqtable[numfreqs - 1].level + 1;
		else //either the other freqset is a higher level, or they're equal
			newfreq.level = freqtable[numfreqs - 2].level + 1;

		//find node number for left node
		for(i = 0; i < numnodes; i++)
			if((hufftable[i].lname == freqtable[numfreqs - 1].node.lname) && (hufftable[i].rname == freqtable[numfreqs - 1].node.rname))
				break;
		if(i == numnodes)   //this can't happen
		{
			printf("Fatal Error; cannot find node to link to. Shutting down...\n");
			return 1;
		}

		newfreq.node.lname = (unsigned short)i;

		//find node number for right node
		for(i = 0; i < numnodes; i++)
			if((hufftable[i].lname == freqtable[numfreqs - 2].node.lname) && (hufftable[i].rname == freqtable[numfreqs - 2].node.rname))
				break;
		if(i == numnodes)   //this can't happen
		{
			printf("Fatal Error; cannot find node to link to. Shutting down...\n");
			return 1;
		}

		newfreq.node.rname = (unsigned short)i;

		//insert new freqency
		freqtable[numfreqs - 2].node.lname = newfreq.node.lname;
		freqtable[numfreqs - 2].node.rname = newfreq.node.rname;
		freqtable[numfreqs - 2].level = newfreq.level;
		freqtable[numfreqs - 2].numoccur = newfreq.numoccur;

		//add node to hufftable
		hufftable[numnodes].lname = newfreq.node.lname;
		hufftable[numnodes].rname = newfreq.node.rname;
		numnodes++;

		numfreqs--;
	}

	//create encoded 'text' file
	enctable = fopen("fe8e.tbl", "wb");
	if(!enctable)
	{
		printf("Could not open target Encoded 'Text' file, fe8e.tbl. Shutting down...\n");
		return 1;
	}
	
	//write out Huffman Table
	fwrite(hufftable, sizeof(NODE), numnodes, enctable);

	fclose(enctable);

	enctable = fopen("fe8e.tbl", "rb");
	createtable(hufftable, numnodes);

	fclose(enctable);

	return 0;
}


/* Comparison function for QuickSort */
int freqcomp(const void* freq1, const void* freq2)
{
    if((((FREQUENCY*)freq2)->numoccur - ((FREQUENCY*)freq1)->numoccur) != 0)
		return (((FREQUENCY*)freq2)->numoccur - ((FREQUENCY*)freq1)->numoccur);
	//frequencies are equal, try by level
	//if levels are equal, we've got nothing. Quicksort can arbitrate.
	return (((FREQUENCY*)freq2)->level - ((FREQUENCY*)freq1)->level);
}

void createtable(NODE *hufftable, int numnodes)
{
	char  bitpattern[100];
	FILE *table;
	int   bplen;
	int   curnode = 0;
	int   findnode = 0;
	int   i;
	int   longest = 0;
	int   searching;
	int   testleft;
	int   testnode;
	int   testright;

	printf("Building Bit Pattern Table, please be patient...\n");

	table = fopen("fe8e.bit", "wb");
	if(!table)
		return;

    curnode = findnode;
	memset(bitpattern, '\0', 100);

	while(findnode < NUM_CHARS)
	{
		if(curnode == findnode) // print character we're looking for
		{
			fprintf(table, "%c%c", hufftable[findnode].lname & 0x00FF, (hufftable[findnode].lname & 0xFF00) >> 8);
		}
		searching = 1;
		testnode = NUM_CHARS - 1;
		//find pointer to this node
		while(searching)
		{
			testnode++;

			if(testnode == numnodes)
			{
				testnode = 0;
				break;
			}

			testleft = hufftable[testnode].lname;
			testright = hufftable[testnode].rname;

			if(testleft == curnode)
				strcat(bitpattern, "0");
			else if(testright == curnode)
				strcat(bitpattern, "1");
			if((testleft == curnode) || (testright == curnode))
				searching = 0;

		}
		if(testnode == 0) //finished that character
		{
			bplen = (int)strlen(bitpattern);
			if(bplen > longest)
				longest = bplen;
			for(i = 0; i < 100; i++)
			{
				if(i < bplen)
					fprintf(table, "%c", bitpattern[bplen - i - 1]);
				else
					fputc('\0', table);
			}
			findnode++;
			if((findnode % 50) == 0)
				printf(".");
		}
		else
		{
            curnode = testnode;
			continue;
		}
		curnode = findnode;
		memset(bitpattern, '\0', 100);
	}

	printf("\n\tLongest pattern: %d bits\n", longest);

	fclose(table);
}